@@ -34,7 +34,6 @@
 #include <ell/io.h>
 #include <ell/ringbuf.h>
 #include <ell/log.h>
-#include <ell/plugin.h>
 #include <ell/checksum.h>
 #include <ell/settings.h>
 #include <ell/hwdb.h>
@@ -46,7 +45,6 @@
 #include <ell/tls.h>
 #include <ell/uuid.h>
 #include <ell/key.h>
-#include <ell/pkcs5.h>
 #include <ell/file.h>
 #include <ell/dir.h>
 #include <ell/net.h>
@@ -57,9 +55,12 @@
 #include <ell/dbus-service.h>
 #include <ell/dbus-client.h>
 #include <ell/dhcp.h>
+#include <ell/dhcp6.h>
+#include <ell/icmp6.h>
 #include <ell/cert.h>
 #include <ell/ecc.h>
 #include <ell/ecdh.h>
 #include <ell/time.h>
 #include <ell/gpio.h>
 #include <ell/path.h>
+#include <ell/acd.h>
