@@ -25,8 +25,6 @@
 #endif
 
 #define _GNU_SOURCE
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <sys/mman.h>
@@ -41,7 +39,6 @@
 #include "base64.h"
 #include "utf8.h"
 #include "asn1-private.h"
-#include "pkcs5-private.h"
 #include "cipher.h"
 #include "cert-private.h"
 #include "missing.h"
@@ -193,17 +190,19 @@ const char *pem_next(const void *buf, size_t buf_len, char **type_label,
 	return NULL;
 }
 
-static uint8_t *pem_load_buffer(const void *buf, size_t buf_len,
+uint8_t *pem_load_buffer(const void *buf, size_t buf_len,
 				char **out_type_label, size_t *out_len,
-				char **out_headers)
+				char **out_headers, const char **out_endp)
 {
 	size_t base64_len;
 	const char *base64;
 	char *label;
+	const char *headers = NULL;
+	size_t headers_len;
 	uint8_t *ret;
 
 	base64 = pem_next(buf, buf_len, &label, &base64_len,
-				NULL, false);
+				out_endp, false);
 	if (!base64)
 		return NULL;
 
@@ -220,7 +219,7 @@ static uint8_t *pem_load_buffer(const void *buf, size_t buf_len,
 
 		if (!(end = memmem(start, base64_len, "\n\n", 2)) &&
 				!(end = memmem(start, base64_len, "\n\r\n", 3)))
-			return NULL;
+			goto err;
 
 		/* Check that each header line has a key and a colon */
 		while (start < end) {
@@ -228,32 +227,40 @@ static uint8_t *pem_load_buffer(const void *buf, size_t buf_len,
 			const char *colon = memchr(start, ':', lf - start);
 
 			if (!colon)
-				return NULL;
+				goto err;
 
 			for (; start < colon; start++)
 				if (l_ascii_isalnum(*start))
 					break;
 
 			if (start == colon)
-				return NULL;
+				goto err;
 
 			start = lf + 1;
 		}
 
-		if (out_headers)
-			*out_headers = l_strndup(base64, end - base64);
+		headers = base64;
+		headers_len = end - base64;
 
-		base64_len -= end + 2 - base64;
+		base64_len -= headers_len + 2;
 		base64 = end + 2;
-	} else if (out_headers)
-		*out_headers = NULL;
+	}
 
 	ret = l_base64_decode(base64, base64_len, out_len);
 	if (ret) {
 		*out_type_label = label;
+
+		if (out_headers) {
+			if (headers)
+				*out_headers = l_strndup(headers, headers_len);
+			else
+				*out_headers = NULL;
+		}
+
 		return ret;
 	}
 
+err:
 	l_free(label);
 
 	return NULL;
@@ -262,16 +269,10 @@ static uint8_t *pem_load_buffer(const void *buf, size_t buf_len,
 LIB_EXPORT uint8_t *l_pem_load_buffer(const void *buf, size_t buf_len,
 					char **type_label, size_t *out_len)
 {
-	return pem_load_buffer(buf, buf_len, type_label, out_len, NULL);
+	return pem_load_buffer(buf, buf_len, type_label, out_len, NULL, NULL);
 }
 
-struct pem_file_info {
-	int fd;
-	struct stat st;
-	uint8_t *data;
-};
-
-static int pem_file_open(struct pem_file_info *info, const char *filename)
+int pem_file_open(struct pem_file_info *info, const char *filename)
 {
 	info->fd = open(filename, O_RDONLY);
 	if (info->fd < 0)
@@ -296,7 +297,7 @@ static int pem_file_open(struct pem_file_info *info, const char *filename)
 	return 0;
 }
 
-static void pem_file_close(struct pem_file_info *info)
+void pem_file_close(struct pem_file_info *info)
 {
 	munmap(info->data, info->st.st_size);
 	close(info->fd);
@@ -315,7 +316,8 @@ static uint8_t *pem_load_file(const char *filename, char **out_type_label,
 		return NULL;
 
 	result = pem_load_buffer(file.data, file.st.st_size,
-					out_type_label, out_len, out_headers);
+					out_type_label, out_len, out_headers,
+					NULL);
 	pem_file_close(&file);
 	return result;
 }
@@ -425,6 +427,7 @@ LIB_EXPORT struct l_queue *l_pem_load_certificate_list_from_data(
 		struct l_cert *cert;
 		const char *base64;
 		size_t base64_len;
+		bool is_certificate;
 
 		base64 = pem_next(ptr, end - ptr, &label,
 					&base64_len, &ptr, false);
@@ -436,17 +439,16 @@ LIB_EXPORT struct l_queue *l_pem_load_certificate_list_from_data(
 			goto error;
 		}
 
-		der = l_base64_decode(base64, base64_len, &der_len);
-
-		if (!der || strcmp(label, "CERTIFICATE")) {
-			if (der)
-				l_free(label);
-			l_free(der);
-
-			goto error;
-		}
-
+		is_certificate = !strcmp(label, "CERTIFICATE");
 		l_free(label);
+
+		if (!is_certificate)
+			goto error;
+
+		der = l_base64_decode(base64, base64_len, &der_len);
+		if (!der)
+			goto error;
+
 		cert = l_cert_new_from_der(der, der_len);
 		l_free(der);
 
@@ -646,95 +648,9 @@ cleanup:
 	return cipher;
 }
 
-struct l_key *pem_key_from_pkcs8_private_key_info(const uint8_t *der,
-							size_t der_len)
-{
-	return l_key_new(L_KEY_RSA, der, der_len);
-}
-
-struct l_key *pem_key_from_pkcs8_encrypted_private_key_info(const uint8_t *der,
-							size_t der_len,
-							const char *passphrase)
-{
-	const uint8_t *key_info, *alg_id, *data;
-	uint8_t tag;
-	size_t key_info_len, alg_id_len, data_len, tmp_len;
-	struct l_cipher *alg;
-	uint8_t *decrypted;
-	int i;
-	struct l_key *pkey;
-	bool r;
-	bool is_block;
-	size_t decrypted_len;
-
-	/* Technically this is BER, not limited to DER */
-	key_info = asn1_der_find_elem(der, der_len, 0, &tag, &key_info_len);
-	if (!key_info || tag != ASN1_ID_SEQUENCE)
-		return NULL;
-
-	alg_id = asn1_der_find_elem(key_info, key_info_len, 0, &tag,
-					&alg_id_len);
-	if (!alg_id || tag != ASN1_ID_SEQUENCE)
-		return NULL;
-
-	data = asn1_der_find_elem(key_info, key_info_len, 1, &tag, &data_len);
-	if (!data || tag != ASN1_ID_OCTET_STRING || data_len < 8 ||
-			(data_len & 7) != 0)
-		return NULL;
-
-	if (asn1_der_find_elem(der, der_len, 2, &tag, &tmp_len))
-		return NULL;
-
-	alg = pkcs5_cipher_from_alg_id(alg_id, alg_id_len, passphrase,
-					&is_block);
-	if (!alg)
-		return NULL;
-
-	decrypted = l_malloc(data_len);
-
-	r = l_cipher_decrypt(alg, data, decrypted, data_len);
-	l_cipher_free(alg);
-
-	if (!r) {
-		l_free(decrypted);
-		return NULL;
-	}
-
-	decrypted_len = data_len;
-
-	if (is_block) {
-		/*
-		 * For block ciphers strip padding as defined in RFC8018
-		 * (for PKCS#5 v1) or RFC1423 / RFC5652 (for v2).
-		 */
-		pkey = NULL;
-
-		if (decrypted[data_len - 1] >= data_len ||
-				decrypted[data_len - 1] > 16)
-			goto cleanup;
-
-		for (i = 1; i < decrypted[data_len - 1]; i++)
-			if (decrypted[data_len - 1 - i] !=
-					decrypted[data_len - 1])
-				goto cleanup;
-
-		decrypted_len -= decrypted[data_len - 1];
-	}
-
-	pkey = pem_key_from_pkcs8_private_key_info(decrypted, decrypted_len);
-
-cleanup:
-	explicit_bzero(decrypted, data_len);
-	l_free(decrypted);
-	return pkey;
-}
-
-static struct l_key *pem_load_private_key(uint8_t *content,
-						size_t len,
-						char *label,
-						const char *passphrase,
-						char *headers,
-						bool *encrypted)
+struct l_key *pem_load_private_key(uint8_t *content, size_t len, char *label,
+					const char *passphrase, char *headers,
+					bool *encrypted)
 {
 	struct l_key *pkey;
 
@@ -749,7 +665,7 @@ static struct l_key *pem_load_private_key(uint8_t *content,
 		if (headers)
 			goto err;
 
-		pkey = pem_key_from_pkcs8_private_key_info(content, len);
+		pkey = cert_key_from_pkcs8_private_key_info(content, len);
 		goto done;
 	}
 
@@ -771,7 +687,7 @@ static struct l_key *pem_load_private_key(uint8_t *content,
 		if (headers)
 			goto err;
 
-		pkey = pem_key_from_pkcs8_encrypted_private_key_info(content,
+		pkey = cert_key_from_pkcs8_encrypted_private_key_info(content,
 								len,
 								passphrase);
 		goto done;
@@ -784,29 +700,9 @@ static struct l_key *pem_load_private_key(uint8_t *content,
 	 * PrivateKeyInfo for the pkcs8-key-parser module.
 	 */
 	if (!strcmp(label, "RSA PRIVATE KEY")) {
-		const uint8_t *data;
-		uint8_t tag;
-		size_t data_len;
-		const uint8_t *key_data;
-		size_t key_data_len;
-		int i;
-		uint8_t *private_key;
-		size_t private_key_len;
-		uint8_t *one_asymmetric_key;
-		uint8_t *ptr;
 		const char *dekalgid;
 		const char *dekparameters;
 
-		static const uint8_t version0[] = {
-			ASN1_ID_INTEGER, 0x01, 0x00
-		};
-		static const uint8_t pkcs1_rsa_encryption[] = {
-			ASN1_ID_SEQUENCE, 0x0d,
-			ASN1_ID_OID, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d,
-			0x01, 0x01, 0x01,
-			ASN1_ID_NULL, 0x00,
-		};
-
 		/*
 		 * "openssl rsa ..." can produce encrypted PKCS#1-formatted
 		 * keys.  These are incompatible with RFC7468 parsing because
@@ -821,8 +717,8 @@ static struct l_key *pem_load_private_key(uint8_t *content,
 		if (dekalgid) {
 			struct l_cipher *alg;
 			bool r;
-			int i;
 			size_t block_len;
+			uint8_t pad;
 
 			if (encrypted)
 				*encrypted = true;
@@ -835,8 +731,10 @@ static struct l_key *pem_load_private_key(uint8_t *content,
 			if (!alg)
 				goto err;
 
-			if (len % block_len || !len)
+			if (len % block_len || !len) {
+				l_cipher_free(alg);
 				goto err;
+			}
 
 			r = l_cipher_decrypt(alg, content, content, len);
 			l_cipher_free(alg);
@@ -845,67 +743,17 @@ static struct l_key *pem_load_private_key(uint8_t *content,
 				goto err;
 
 			/* Remove padding like in RFC1423 Section 1.1 */
-			if (content[len - 1] > block_len)
+			pad = content[len - 1];
+			if (pad > block_len || pad == 0)
 				goto err;
 
-			for (i = 1; i < content[len - 1]; i++)
-				if (content[len - 1 - i] != content[len - 1])
-					goto err;
-
-			len -= content[len - 1];
-		}
-
-		/*
-		 * Sanity check that it's a version 0 or 1 RSAPrivateKey
-		 * structure with the 8 integers, if it's not, make a last
-		 * ditch attempt to load it into the kernel directly.
-		 */
-		key_data = asn1_der_find_elem(content, len, 0, &tag,
-						&key_data_len);
-		if (!key_data || tag != ASN1_ID_SEQUENCE)
-			goto err;
-
-		data = asn1_der_find_elem(key_data, key_data_len, 0, &tag,
-						&data_len);
-		if (!data || tag != ASN1_ID_INTEGER || data_len != 1 ||
-				(data[0] != 0x00 && data[0] != 0x01))
-			goto err;
-
-		for (i = 1; i < 9; i++) {
-			data = asn1_der_find_elem(key_data, key_data_len,
-							i, &tag, &data_len);
-			if (!data || tag != ASN1_ID_INTEGER || data_len < 1)
+			if (!l_secure_memeq(content + len - pad, pad - 1U, pad))
 				goto err;
+
+			len -= pad;
 		}
 
-		private_key = l_malloc(10 + len);
-		ptr = private_key;
-		*ptr++ = ASN1_ID_OCTET_STRING;
-		asn1_write_definite_length(&ptr, len);
-		memcpy(ptr, content, len);
-		ptr += len;
-		private_key_len = ptr - private_key;
-
-		one_asymmetric_key = l_malloc(32 + private_key_len);
-		ptr = one_asymmetric_key;
-		*ptr++ = ASN1_ID_SEQUENCE;
-		asn1_write_definite_length(&ptr,
-						sizeof(version0) +
-						sizeof(pkcs1_rsa_encryption) +
-						private_key_len);
-		memcpy(ptr, version0, sizeof(version0));
-		ptr += sizeof(version0);
-		memcpy(ptr, pkcs1_rsa_encryption, sizeof(pkcs1_rsa_encryption));
-		ptr += sizeof(pkcs1_rsa_encryption);
-		memcpy(ptr, private_key, private_key_len);
-		ptr += private_key_len;
-		explicit_bzero(private_key, private_key_len);
-		l_free(private_key);
-
-		pkey = pem_key_from_pkcs8_private_key_info(one_asymmetric_key,
-						ptr - one_asymmetric_key);
-		explicit_bzero(one_asymmetric_key, ptr - one_asymmetric_key);
-		l_free(one_asymmetric_key);
+		pkey = cert_key_from_pkcs1_rsa_private_key(content, len);
 		goto done;
 	}
 
@@ -933,7 +781,7 @@ LIB_EXPORT struct l_key *l_pem_load_private_key_from_data(const void *buf,
 	if (encrypted)
 		*encrypted = false;
 
-	content = pem_load_buffer(buf, buf_len, &label, &len, &headers);
+	content = pem_load_buffer(buf, buf_len, &label, &len, &headers, NULL);
 
 	if (!content)
 		return NULL;
@@ -955,6 +803,10 @@ LIB_EXPORT struct l_key *l_pem_load_private_key_from_data(const void *buf,
  * success case and on error when NULL is returned.  This can be used to
  * check if a passphrase is required without prior information.
  *
+ * The passphrase, if given, must have been validated as UTF-8 unless the
+ * caller knows that PKCS#12 encryption algorithms are not used.
+ * Use l_utf8_validate.
+ *
  * Returns: An l_key object to be freed with an l_key_free* function,
  * or NULL.
  **/
